unit uSignalIn;

interface

uses
  Classes, IniFiles, uWaveControl, uWaveIn, uCom, uFFT;
const
  cnPolN = 15;
type
  TComArray = array[0..MaxInt div SizeOf(Com) - 1] of Com;
  pComMask = ^TComArray;
  TPol = array [0..cnPolN] of Extended;
  TMat = array [0..cnPolN, 0..cnPolN] of Extended;
  enSignalMode = (smAdjustment, smCalibration, smAcquisition, smRaw);
  TBoolProc = procedure(aValue: Boolean) of object;
  TRealProc = procedure(aValue: Double) of object;
  TSignalIn = class(TSyncIn)
  private
    FMode: enSignalMode;
    FHarmId, FHarmdWin: Integer;
    FHarmVal: array [0..1] of Com;
    FStrProc: TGetStrProc;
    FOnSaturation: TBoolProc;
    FSetAmp: TRealProc;
    FChanCal: array of Com;
    FFirstCall: Boolean;
    FMat: TMat;
    FCalibrData: record
      Sat: array [0..1] of Double;
      Pol: array [0..1] of TPol;
    end;
    procedure GetMat;
    function GetPol(aID: Integer): TPol;
    procedure GetChanCal;
  protected
    procedure AsyncProc(aData: TWaveData); override;
  public
    constructor Create;
    procedure DoAdjustement(aFreq: Double; aStrProc: TGetStrProc);
    procedure DoCalibration(aMinFreq, aMaxFreq: Double);
    procedure DoAcquisition(aRawMode: Boolean);
    procedure ProcZLCRMeter(aData: TWaveData; aMinFreq, aMaxFreq, aRref: Double;
                            aSwap: Boolean; aCalcLCR: Boolean; aStrProc: TGetStrProc);
    procedure Transfer(aData: TWaveData; aMinFreq, aMaxFreq: Double; aSwap: Boolean);
    property OnSaturation: TBoolProc write FOnSaturation;
    property SetAmp: TRealProc write FSetAmp;
    procedure LoadIni(aFile: TIniFile);
    procedure SaveIni(aFile: TIniFile);
  end;

implementation

uses
  Sysutils, Math;

const
  csSection = 'Calibration';

procedure CholDec(var A : TMat);
var
  i, j, k: Integer;
  Ajj, V : Extended;
begin
  for j:= 0 to High(A) do begin
    V:= 0;
    for k := 0 to j-1 do
      V:= V + Sqr(A[k,j]);
    Ajj:= A[j,j] - V;
    Ajj:= Sqrt(Ajj);
    A[j,j]:= Ajj;
    if j < High(A) then begin
      for i:= j+1 to High(A) do begin
        V:= 0;
        for k:= 0 to j-1 do
          V:= V + A[k,i]*A[k,j];
        A[j,i]:= A[j,i] - V;
      end;
      V:= 1/Ajj;
      for k:= j+1 to High(A) do
        A[j,k]:= A[j,k] * V;
    end;
  end;
end;

procedure CholRes(A: TMat; var X: TPol);
var
  i, j: Integer;
  V: Double;
begin
  X[0]:= X[0] / A[0,0];
  for i:= 1 to High(A) do begin
    V:= 0;
    for j:= 0 to i-1 do
      V:= V + A[j,i] * X[j];
    X[i]:= (X[i] - V) / A[i,i];
  end;
  X[High(X)]:= X[High(X)] / A[High(A),High(A)];
  for i:= High(X) - 1 downto 0 do begin
    V:= 0;
    for j:= i+1 to High(A) do
      V:= V + A[i,j] * X[j];
    X[i]:= (X[i] - V) / A[i,i];
  end;
end;

function PolRes(aPol: TPol; aX: Double): Double;
var
  i: Integer;
begin
  Result:= aPol[High(aPol)];
  for i:= High(aPol) - 1 downto 0 do
    Result:= Result * aX + aPol[i];
end;

constructor TSignalIn.Create;
var
  i: Integer;
begin
  inherited;
  for i:= 0 to 1 do
    FHarmVal[i]:= Com1;
  FMode:= smAcquisition;
  FOnSaturation:= nil;
  FStrProc:= nil;
  for i:= 0 to 1 do with FCalibrData do begin
    Sat[i]:= 0.9;
    FillChar(Pol[i], SizeOf(Pol[i]), 0);
  end;
  FCalibrData.Pol[0,0]:= 1;
end;

procedure TSignalIn.GetMat;
var
  i, j: Integer;
  ax2, axn: Extended;
begin
  FillChar(FMat, SizeOf(FMat), 0);
  FMat[0,0]:= 2*FHarmdWin + 1;
  for i:= 1 to FHarmdWin do begin
    axn:= 2;
    ax2:= Sqr(i);
    for j:= 1 to High(FMat) do begin
      axn:= axn * ax2;
      FMat[j,j]:= FMat[j,j] + axn;
    end;
  end;
  for i:= 1 to High(FMat) - 1 do begin
    j:= i;
    repeat
      Inc(j);
      FMat[2*i-j,j]:= FMat[i,i];
    until (j = 2*i) or (j = High(FMat));
  end;
  CholDec(FMat);
end;

function TSignalIn.GetPol(aID: Integer): TPol;
var
  i, j: Integer;
  axn, ayp, aym, ay: Extended;
  aPol, aX0N: TPol;
  aBaseFreq: Extended;
begin
  aBaseFreq:= Sampling / DataCount;
  FillChar(aPol, SizeOf(aPol), 0);
  aX0N:= aPol;
  aPol[0]:= FChanCal[FHarmID].u[aId];
  for i:= 1 to FHarmdWin do begin
    ay:= FChanCal[FHarmID-i].u[aId];
    ayp:= FChanCal[FHarmID+i].u[aId] + ay;
    aym:= FChanCal[FHarmID+i].u[aId] - ay;
    aPol[0]:= aPol[0] + ayp;
    axn:= 1;
    for j:= 1 to High(aPol) do begin
      if (j mod 2) = 0 then
        ay:= ayp
      else
        ay:= aym;
      axn:= axn * i;
      aPol[j]:= aPol[j] + ay * axn;
    end;
  end;
  CholRes(FMat, aPol);
  Result:= aPol;
  aX0N[0]:= 1;
  for i:= 1 to High(Result) do begin
    aX0N[i]:= 1;
    for j:= i - 1 downto 0 do begin
      aX0N[j]:= -FHarmId * aX0N[j];
      if j > 0 then
        aX0N[j]:= aX0N[j] + aX0N[j-1];
      Result[j]:= Result[j] + aX0N[j]*aPol[i];
    end;
  end;
  ay:= 1;
  for i:= 1 to High(Result) do begin
    ay:= ay / aBaseFreq;
    Result[i]:= Result[i] * ay;
  end;
end;

procedure TSignalIn.GetChanCal;
var
  i, j: Integer;
  aBaseFreq: Extended;
begin
  aBaseFreq:= Sampling / DataCount;
  for i:= 1 to High(FChanCal) do
    for j:= 0 to 1 do
      FChanCal[i].u[j]:= PolRes(FCalibrData.Pol[j], i*aBaseFreq);
end;

procedure TSignalIn.AsyncProc(aData: TWaveData);
var
  i, j: Integer;
  aNewHarmVal: array [0..1] of Com;
  aVal, aNorm: Double;
  aSt: String;
  aRat: Com;
  aMaxAmpl: array [0..1] of Double;
  aSp0, aSp1: pComMask;
  aSaturation: Boolean;
begin
  for i:= 0 to 1 do begin
    aMaxAmpl[i]:= Abs(aData[i,0]);
    for j:= 1 to High(aData[i]) do
      if Abs(aData[i,j]) > aMaxAmpl[i] then
        aMaxAmpl[i]:= Abs(aData[i,j]);
    WaveFFT(@aData[i,0], Length(aData[i]), True);
  end;
  aSaturation:= (aMaxAmpl[0] >= FCalibrData.Sat[0]) or
                (aMaxAmpl[1] >= FCalibrData.Sat[1]);
  aSp0:= @aData[0,0];
  aSp1:= @aData[1,0];

  case FMode of

    smAdjustment:
      begin
        aSt:= 'Quality:';
        for i:= 0 to 1 do begin
          aNewHarmVal[i]:= cCom(aData[i, 2*FHarmId]);
          aVal:= 0;
          j:= FHarmId*2;
          while j < Length(aData[i]) div 2 - 1 do begin
            aVal:= aVal + cMod2(cCom(aData[i, 2*j]));
            j:= j + FHarmId;
          end;
          aSt:= aSt + ' ' + FloatToStrF(
            10 * Log10(cMod2(aNewHarmVal[i])/aVal), ffFixed, 0, 0);
        end;
        if not FFirstCall then begin
          aVal:= 0;
          for i:= 0 to 1 do begin
            if FCalibrData.Sat[i] < aMaxAmpl[i] then
              FCalibrData.Sat[i]:= aMaxAmpl[i];
            aRat:= cDiv(aNewHarmVal[i], FHarmVal[i]);
            if aRat.x <> 0 then
              aVal:= aVal + aRat.y / aRat.x
            else
              aVal:= aVal - 100;
          end;
           aSt:= aSt +  Chr(13) + Chr(10) + 'Async: ' +
             FloatToStrF(aVal, ffFixed, 0, 3);
        end;
        for i:= 0 to 1 do
          FHarmVal[i]:= aNewHarmVal[i];
        if Assigned(FStrProc) then
          FStrProc(aSt);
      end;

    smCalibration:
      //if not FFirstCall then
      begin
        for i:= 1 to Length(aData[0]) div 2 - 1 do begin
          aRat:= cDiv(aSp0[i], aSp1[i]);
          //aRat:= cAmPh(aRat);
          aSp0[i]:= aRat;
          FChanCal[i]:= aRat;
        end;
        for i:= 0 to 1 do
          FCalibrData.Pol[i]:= GetPol(i);
          GetChanCal;
        for i:= 1 to Length(aData[0]) div 2 - 1 do begin
          aSp1[i]:= FChanCal[i];
        end;
        {
        for i:= 1 to Length(aData[0]) div 2 - 1 do begin
          aSp1[i]:= cSub(FChanCal[i], aSp0[i]);
          aSp0[i]:= aSp1[i];
        end;
            }
        if aSaturation then
          aSt:= 'Wait...'
        else
          aSt:= 'Press Stop'+  Chr(13) + Chr(10)
              + 'to finish.';
        if Assigned(FStrProc) then
          FStrProc(aSt);
     end;

    smAcquisition:
      begin
        aRat:= Com0;
        for i:= 0 to 1 do if FCalibrData.Sat[i] > 0 then
          aRat.u[i]:= aMaxAmpl[i] / FCalibrData.Sat[i];
        if aRat.x < aRat.y then
          aRat.x:= aRat.y;
        if (aRat.x > 0) and (aRat.x < 0.7) and Assigned(FSetAmp) then
             FSetAmp(0.7/aRat.x);
        for i:= 1 to Length(aData[0]) div 2 - 1 do begin
          aSp1[i]:= cMul(aSp1[i], FChanCal[i]);
        end;
      end;
      
    smRaw:
      begin
        aRat:= Com0;
        for i:= 0 to 1 do if FCalibrData.Sat[i] > 0 then
          aRat.u[i]:= aMaxAmpl[i] / FCalibrData.Sat[i];
        if aRat.x < aRat.y then
          aRat.x:= aRat.y;
        if (aRat.x > 0) and (aRat.x < 0.7) and Assigned(FSetAmp) then
             FSetAmp(0.7/aRat.x);
        aNorm:= 100/(Length(aData[0]) div 2);
        for i:= 1 to Length(aData[0]) div 2 - 1 do begin
          aSp0[i]:= cMuR(aSp0[i], aNorm);
          aSp1[i]:= cMuR(aSp1[i], aNorm);
        end;
      end;
  end;

  if (FMode <> smAdjustment) then begin
    if Assigned(FOnSaturation) then
      FOnSaturation(aSaturation);
    if Assigned(FSetAmp) and aSaturation then
      FSetAmp(0.9);
  end;
  FFirstCall:= False;
end;

procedure TSignalIn.DoAdjustement(aFreq: Double; aStrProc: TGetStrProc);
begin
  FHarmId:= FWaveIn.GetHarmonic(aFreq, DataCount);
  FStrProc:= aStrProc;
  FMode:= smAdjustment;
  FFirstCall:= True;
  with FCalibrData do
    FillChar(Sat, SizeOf(Sat), 0);
end;

procedure TSignalIn.DoCalibration(aMinFreq, aMaxFreq: Double);
var
  aMin, aMax: Integer;
begin
  FMode:= smCalibration;
  aMin:= FWaveIn.GetHarmonic(aMinFreq, DataCount);
  aMax:= FWaveIn.GetHarmonic(aMaxFreq, DataCount);
  FHarmId:= (aMin + aMax) div 2;
  FHarmdWin:= FHarmId - aMin;
  SetLength(FChanCal, DataCount div 2);
  GetMat;
end;

procedure TSignalIn.DoAcquisition(aRawMode: Boolean);
begin
  if aRawMode then
    FMode:= smRaw
  else begin
    FMode:= smAcquisition;
    SetLength(FChanCal, DataCount div 2);
    GetChanCal;
  end;
end;

function ValToStr(aValue: Extended): String;
const cPrecision = 4;
      cDigits = 0;
begin
  if Abs(aValue) > 1E3 then
  if Abs(aValue) > 1E6 then
    Result:= FloatToStrF(aValue * 1E-6, ffGeneral, cPrecision, cDigits) + ' M'
  else
    Result:= FloatToStrF(aValue * 1E-3, ffGeneral, cPrecision, cDigits) + ' k'
  else
  if Abs(aValue) < 1.0 then
  if Abs(aValue) < 1E-3 then
  if Abs(aValue) < 1E-6 then
  if Abs(aValue) < 1E-9 then
    Result:= FloatToStrF(aValue * 1E12, ffGeneral, cPrecision, cDigits) + ' p'
  else
    Result:= FloatToStrF(aValue * 1E9,  ffGeneral, cPrecision, cDigits) + ' n'
  else
    Result:= FloatToStrF(aValue * 1E6,  ffGeneral, cPrecision, cDigits) + ' '
  else
    Result:= FloatToStrF(aValue * 1E3,  ffGeneral, cPrecision, cDigits) + ' m'
  else
    Result:= FloatToStrF(aValue,        ffGeneral, cPrecision, cDigits) + ' ';
end;

procedure TSignalIn.ProcZLCRMeter(aData: TWaveData;
  aMinFreq, aMaxFreq, aRref: Double;
  aSwap: Boolean; aCalcLCR: Boolean; aStrProc: TGetStrProc);
var
  i, aMin, aMax: Integer;
  aL, aC, aR,
  aBaseFreq, aOmega, aTg: Double;
  aIn0, aIn1, aOut0, aOut1: pComMask;
  aSt: String;
begin
  if aSwap then begin
    aIn1:= @aData[0,0];
    aIn0:= @aData[1,0];
  end else begin
    aIn0:= @aData[0,0];
    aIn1:= @aData[1,0];
  end;
  aOut0:= @aData[0,0];
  aOut1:= @aData[1,0];
  aMin:= FWaveIn.GetHarmonic(aMinFreq, DataCount);
  aMax:= FWaveIn.GetHarmonic(aMaxFreq, DataCount);
  aBaseFreq:= 2*Pi*Sampling / DataCount;
  aL:= 0;
  aC:= 0;
  aR:= 0;
  aTg:= 0;
  for i:= aMin to aMax do begin
    aOut0[i]:= cDiv(aIn0[i], aIn1[i]);
    aOut0[i].x:= aRref*(aOut0[i].x - 1);
    aOut0[i].y:= -aRref*aOut0[i].y;
    if aCalcLCR then begin
      aOmega:= aBaseFreq * i;
      aR:= aR + aOut0[i].x;
      aL:= aL + aOut0[i].y / aOmega;
      aC:= aC + -1 / (aOut0[i].y * aOmega);
      aTg:= aTg + aOut0[i].x / Abs(aOut0[i].y);
    end;
  end;
  if not aCalcLCR then
    Exit;
  i:= (aMax-aMin+1);
  aR:= aR / i;
  aL:= aL / i;
  aC:= aC / i;
  aTg:= aTg / i;
  if aL > 0 then
    aSt:= 'Lx= ' + ValToStr(aL) + 'H'
  else
  if aC > 0 then
    aSt:= 'Cx= ' + ValToStr(aC) + 'F'
  else
    aSt:= 'Im= ?';
  aSt:= aSt + Chr(13) + Chr(10) +
    'R= ' + ValToStr(aR) + 'Ohm';
  aSt:= aSt + Chr(13) + Chr(10) +
    'tg= ' + FloatToStrF(aTg, ffExponent, 2, 2);
  for i:= aMin to aMax do begin
    aOmega:= aBaseFreq * i;
    aOut1[i].x:= aR;
    if aL > 0 then
      aOut1[i].y:= aOmega * aL
    else
      aOut1[i].y:= -1 / (aOmega * aC);
  end;
  if Assigned(aStrProc) then
    aStrProc(aSt);
end;

procedure TSignalIn.Transfer(aData: TWaveData; aMinFreq, aMaxFreq: Double;
                             aSwap: Boolean);
var
  i, aMin, aMax: Integer;
  aIn0, aIn1, aOut0: pComMask;
begin
  if aSwap then begin
    aIn1:= @aData[0,0];
    aIn0:= @aData[1,0];
  end else begin
    aIn0:= @aData[0,0];
    aIn1:= @aData[1,0];
  end;
  aOut0:= @aData[0,0];
  aMin:= FWaveIn.GetHarmonic(aMinFreq, DataCount);
  aMax:= FWaveIn.GetHarmonic(aMaxFreq, DataCount);
  for i:= aMin to aMax do
    aOut0[i]:= cDiv(aIn1[i], aIn0[i]);
end;

procedure TSignalIn.LoadIni(aFile: TIniFile);
var
  i, j: Integer;
begin
  for i:= 0 to 1 do
    FCalibrData.Sat[i]:= aFile.ReadFloat(csSection,
      'Saturation' + IntToStr(i), 0.9);
  for i:= 0 to cnPolN do
     for j:= 0 to 1 do
       FCalibrData.Pol[j,i]:= aFile.ReadFloat(csSection,
         'Polynom' + IntToStr(i) + IntToStr(j), FCalibrData.Pol[j,i]);
  GetChanCal;
end;

procedure TSignalIn.SaveIni(aFile: TIniFile);
var
  i, j: Integer;
begin
  for i:= 0 to 1 do
    aFile.WriteFloat(csSection, 'Saturation' + IntToStr(i), FCalibrData.Sat[i]);
  for i:= 0 to cnPolN do
     for j:= 0 to 1 do
       aFile.WriteFloat(csSection,
         'Polynom' + IntToStr(i) + IntToStr(j), FCalibrData.Pol[j,i]);
end;

end.
