unit uWaveControl;

interface                                       

uses
  Forms,
  MMSystem, Classes, Windows;
  
const
  {$EXTERNALSYM WAVE_FORMAT_DIRECT}
  WAVE_FORMAT_DIRECT = $0008;
  cnBuffCount = 10;
  cnBuffTime = 0.1; // seconds

type
  enEventType = (etOpen, etData, etReset, etClose);
  TChannelData = array of Double;
  TWaveData = array[0..1] of TChannelData;
  TBufferData = array of array[0..1] of SmallInt;
  TDataMask = array[0..MaxInt div 4 - 1, 0..1] of SmallInt;
  pDataMask = ^TDataMask;
  TBooleanProc = procedure(aBoolean: Boolean) of object;

const
  cnNil2: TWaveData = (nil, nil);

type
  TWaveControl = class(TThread)
  private
    FBufID: Integer;
    procedure SetSampling(aValue: LongWord);
    function GetSampling: LongWord;
    procedure UpdateFormatEx;
    procedure SetDevice(aDevice: LongWord);
  protected
    FDeviceID: LongWord;
    FDeviceHD: Integer;
    FFormatEx: TWaveFormatEx;
    FWHdr: array [0..cnBuffCount-1] of TWaveHdr;
    FPosition: Integer;    
    FEvent: THandle;
    FEventType: enEventType;
    FErrorMes: TGetStrProc;
    procedure AllocBuffers;
    function DoStart: Boolean;
    function DoStop: Boolean;
    procedure Execute; override;
    procedure OnStop; virtual;
    procedure OnDestroy; virtual;
    procedure ExecData(aWaveHdr: pWAVEHDR); virtual; abstract;
    procedure CloseWave; virtual;
    procedure WaveReset; virtual; abstract;
  public
    constructor Create;
    destructor Destroy; override;
    function GetHarmonic(aFreq: Double; aCount: Integer): Integer;    
    property Sampling: LongWord read GetSampling write SetSampling;
    property Device: LongWord read FDeviceID write SetDevice;
    property ErrorMes: TGetStrProc read FErrorMes write FErrorMes;
  end;

procedure SetWaveLength(var aWaveData: TWaveData; aCount: Integer);
procedure MoveWave(aDataSrc, aDataDst: TWaveData);

implementation

uses
  Sysutils;

procedure SetWaveLength(var aWaveData: TWaveData; aCount: Integer);
var
  i: Integer;
begin
  if Length(aWaveData[0]) = aCount then
    Exit;
  for i:= 0 to High(aWaveData) do
    SetLength(aWaveData[i], aCount);
end;

procedure MoveWave(aDataSrc, aDataDst: TWaveData);
var
  i: Integer;
begin
  for i:= 0 to High(TWaveData) do
    Move(aDataSrc[i,0], aDataDst[i,0], Length(aDataSrc[i])*SizeOf(aDataSrc[i,0]));
end;

constructor TWaveControl.Create;
begin
  inherited Create(True);
  FEvent:= CreateEvent(nil, false, false, nil);
  FFormatEx.wFormatTag:= WAVE_FORMAT_PCM;
  FFormatEx.nChannels:= 2;
  FFormatEx.wBitsPerSample:= 16;
  FFormatEx.nSamplesPerSec:= 48000;
  UpdateFormatEx;
  FErrorMes:= nil;
end;

destructor TWaveControl.Destroy;
begin
  if Suspended then
    OnDestroy
  else begin
    FEventType:= etReset;
    WaveReset;
  end;
  inherited;
end;

procedure TWaveControl.CloseWave;
begin
  FEventType:= etClose;
end;

procedure TWaveControl.Execute;
begin
  repeat
    WaitForSingleObject(FEvent, INFINITE);
    case FEventType of
      etOpen: FEventType:= etData;
      etData: while (FWHdr[FBufID].dwFlags and WHDR_INQUEUE) = 0 do begin
                ExecData(@FWHdr[FBufID]);
                if FBufID < High(FWHdr) then
                  FBufID:= FBufID + 1
                else
                  FBufID:= 0;
              end;
      etReset: begin
                 CloseWave;
               end;
      etClose: begin
                 OnStop;
                 if not Terminated then
                   Suspend;
               end;
    end;
    if Terminated then
      Break;
  until False;
  OnDestroy;
end;

procedure TWaveControl.SetDevice(aDevice: LongWord);
begin
  FDeviceID:= aDevice;
end;

procedure TWaveControl.SetSampling(aValue: LongWord);
begin
  FFormatEx.nSamplesPerSec:= aValue;
end;

function TWaveControl.GetSampling: LongWord;
begin
  Result:= FFormatEx.nSamplesPerSec;
end;

procedure TWaveControl.UpdateFormatEx;
begin
  with FFormatEx do begin
    nBlockAlign:= (wBitsPerSample div 8) * nChannels;
    nAvgBytesPerSec:= nSamplesPerSec * nBlockAlign;
    cbSize:= 0;
  end;
end;

procedure TWaveControl.AllocBuffers;
var
  aSI: TSystemInfo;
  aMemSize, aPgCount: LongWord;
  i: Integer;
begin
  UpdateFormatEx;
  GetSystemInfo(aSI);
  with FFormatEx, aSI do begin
    aPgCount:= (Round(nSamplesPerSec*cnBuffTime) * nBlockAlign) div dwPageSize;
    if aPgCount < 1 then
      aPgCount:= 1;
    aMemSize:= aPgCount * dwPageSize;
  end;
  for i:= 0 to High(FWHdr) do begin
    FillChar(FWHdr[i], SizeOf(FWHdr[i]), 0);
    with FWHdr[i] do begin
      dwBufferLength:= aMemSize;
      lpData:= VirtualAlloc(nil, dwBufferLength,
        MEM_RESERVE or MEM_COMMIT, PAGE_READWRITE);
    end;
  end;
end;

procedure TWaveControl.OnStop;
var
  i: Integer;
begin
  for i:= 0 to High(FWHdr) do
    VirtualFree(FWHdr[i].lpData, 0, MEM_RELEASE);
end;

function TWaveControl.DoStart: Boolean;
begin
  Result:= Suspended;
  if not Result then
    Exit;
  Resume;
  AllocBuffers;
  FEventType:= etOpen;
  FBufID:= 0;
end;

function TWaveControl.DoStop: Boolean;
begin
  Result:= not Suspended;
  FEventType:= etReset;
end;

procedure TWaveControl.OnDestroy;
begin
  CloseHandle(FEvent);
end;

function TWaveControl.GetHarmonic(aFreq: Double; aCount: Integer): Integer;
begin
  Result:= Round(aFreq*aCount/Sampling);
  if Result < 1 then
    Result:= 1;
  if Result > aCount div 2 - 1 then
    Result:= aCount div 2 - 1;
end;

end.
