unit uWaveIn;

interface
uses
  MMSystem, uWaveControl,
  Classes, Windows, SysUtils;

type
TWaveProc = procedure(aData: TWaveData) of object;

TWaveIn = class(TWaveControl)
private
  FWaveProc: TWaveProc;
  FNewDataCount: Integer;
  FData: TWaveData;
  FValidData: Boolean;
  FAllowValidation: Boolean;
  FValidationNum: Integer;
  function GetDataCount: Integer;
  procedure SetDataCount(aCount: Integer);
protected
  procedure OnStop; override;
  procedure ExecData(aWaveHdr: pWAVEHDR); override;
  procedure WaveReset; override;
  procedure CloseWave; override;
public
  constructor Create;
  destructor Destroy; override;
  function Start: Boolean;
  procedure Validation(aVal: Boolean);
  procedure Stop;
  property Data: TWaveData read FData;
  property DataCount: Integer read GetDataCount write SetDataCount;
  property WaveProc: TWaveProc write FWaveProc;
end;

TSyncIn = class(TThread)
private
  FEvent: THandle;
  FWaveProc: TWaveProc;
  FData: TWaveData;
  FDateTime: TDateTime;
  FAccumulation: Boolean;
  FAccData: TWaveData;
  FAccInd: Integer;
  FAllowData: Boolean;
  FAccTime: Integer;
  function GetDevice: LongWord;
  procedure SetDevice(aDevice: LongWord);
  function GetSampling: LongWord;
  procedure SetSampling(aSampling: LongWord);
  function GetDataCount: Integer;
  procedure SetDataCount(aCount: Integer);
  procedure SetErrorMes(aErrorMes: TGetStrProc);
  procedure OnDestroy;
  procedure ExecEvent;
  procedure AcquireData(aData: TWaveData);
  procedure SendData(aData: TWaveData);
protected
  FWaveIn: TWaveIn;
  procedure Execute; override;
  procedure OnlineProc(aData: TWaveData); virtual;
  procedure AsyncProc(aData: TWaveData); virtual;
public
  constructor Create; 
  destructor Destroy; override;
  function Start: Boolean;
  procedure Stop;
  procedure AllowData(Sender: TObject);
  procedure Validation(aVal: Boolean);
  procedure SetAccTime(aTime: String);
  property WaveProc: TWaveProc write FWaveProc;
  property Sampling: LongWord read GetSampling write SetSampling;
  property Device: LongWord read GetDevice write SetDevice;
  property DataCount: Integer read GetDataCount write SetDataCount;
  property ErrorMes: TGetStrProc write SetErrorMes;
  property Accumulation: Boolean read FAccumulation write FAccumulation;
end;

implementation

uses Dialogs;

constructor TWaveIn.Create;
begin
  inherited;
  DataCount:= 1 shl 8;
  FWaveProc:= nil;
end;

destructor TWaveIn.Destroy;
begin
  inherited;
end;

function TWaveIn.Start: Boolean;
var
  i: Integer;
  aSampling: Int64;
begin
  Result:= True;
  if not DoStart then
    Exit;
  aSampling:= Sampling;
  FPosition:= -25-aSampling div 1000;
  FValidData:= True;
  if waveInOpen(@FDeviceHD, FDeviceID, @FFormatEx, FEvent, 0,
    CALLBACK_EVENT or WAVE_FORMAT_DIRECT)<> MMSYSERR_NOERROR then begin
      Suspend;
      inherited OnStop;
      Result:= False;
      if Assigned(FErrorMes) then
        FErrorMes('Can not open input device ' + IntToStr(FDeviceID));
      Exit;
    end;
  for i:= 0 to High(FWHdr) do begin
    waveInPrepareHeader(FDeviceHD, @FWHdr[i], SizeOf(FWHdr[i]));
    waveInAddBuffer(FDeviceHD, @FWHdr[i], SizeOf(FWHdr[i]));
  end;
  waveInStart(FDeviceHD);
end;

procedure TWaveIn.Validation(aVal: Boolean);
begin
  if aVal then begin
    FAllowValidation:= True;
    FValidationNum:= 3;
  end else begin
    FValidData:= False;
    FPosition:= 0;
    FAllowValidation:= False;
  end;
end;

procedure TWaveIn.ExecData(aWaveHdr: pWAVEHDR);
const
  cRange = 2 / (High(SmallInt) - Low(SmallInt));
var
  i, j: Integer;
  aMask: pDataMask;
begin
  aMask:= pDataMask(aWaveHdr.lpData);
  if FNewDataCount <> Length(FData[0]) then begin
    SetWaveLength(FData, FNewDataCount);
    if FPosition > High(FData[0]) then
      FPosition:= 0;
  end;
  if FValidData then
    for i:= 0 to (aWaveHdr.dwBytesRecorded div 4) - 1 do begin
      if FPosition < 0 then begin
        Inc(FPosition);
        Continue;
      end;
      for j:= 0 to 1 do
        FData[j, FPosition]:= cRange * aMask[i,j];
      if FPosition = High(FData[0]) then begin
        FPosition:= 0;
        if Assigned(FWaveProc) then
          FWaveProc(FData);
      end else
        Inc(FPosition);
    end
  else
    if FAllowValidation then begin
      Dec(FValidationNum);
      if FValidationNum = 0 then
        FValidData:= True;
    end;  
  waveInAddBuffer(FDeviceHD, aWaveHdr, SizeOf(aWaveHdr^));
end;

procedure TWaveIn.CloseWave;
begin
  waveInStop(FDeviceHD);
  inherited;
  waveInClose(FDeviceHD);
end;

procedure TWaveIn.WaveReset;
begin
  waveInReset(FDeviceHD);
end;

procedure TWaveIn.Stop;
begin
  if not DoStop then
    Exit;
  waveInReset(FDeviceHD);
end;

procedure TWaveIn.OnStop;
var
  i: Integer;
begin
  for i:= 0 to High(FWHdr) do
    waveInUnPrepareHeader(FDeviceHD, @FWHdr[i], SizeOf(FWHdr[i]));
  inherited;
end;

function TWaveIn.GetDataCount: Integer;
begin
  Result:= Length(FData[0]);
end;

procedure TWaveIn.SetDataCount(aCount: Integer);
begin
  FNewDataCount:= aCount;
  if Suspended then
    SetWaveLength(FData, FNewDataCount);
end;

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

constructor TSyncIn.Create;
begin
  inherited Create(True);
  FWaveIn:= TWaveIn.Create;
  FWaveIn.WaveProc:= AcquireData;
  FEvent:= CreateEvent(nil, false, false, nil);
  FDateTime:= 0;
  FAccumulation:= False;
end;

destructor TSyncIn.Destroy;
begin
  FWaveIn.Free;
  if Suspended then
    OnDestroy;
  Terminate;
  PulseEvent(FEvent);
  inherited;
end;

procedure TSyncIn.OnDestroy;
begin
  CloseHandle(FEvent);
end;

function TSyncIn.GetDataCount: Integer;
begin
  Result:= FWaveIn.DataCount;
end;

procedure TSyncIn.SetDataCount(aCount: Integer);
begin
  FWaveIn.DataCount:= aCount;
end;

function TSyncIn.GetDevice: LongWord;
begin
  Result:= FWaveIn.Device;
end;

procedure TSyncIn.SetDevice(aDevice: LongWord);
begin
  FWaveIn.Device:= aDevice;
end;

function TSyncIn.GetSampling: LongWord;
begin
  Result:= FWaveIn.Sampling;
end;

procedure TSyncIn.SetSampling(aSampling: LongWord);
begin
  FWaveIn.Sampling:= aSampling;
end;

procedure TSyncIn.Execute;
begin
  repeat
    WaitForSingleObject(FEvent, INFINITE);
    if Terminated then
      Break;
    Synchronize(ExecEvent);
  until False;
  OnDestroy;
end;

function TSyncIn.Start: Boolean;
begin
  Result:= True;
  if not Suspended then
    Exit;
  Result:= False;  
  FAccInd:= 0;
  FAllowData:= True;
  Resume;
  if not FWaveIn.Start then
    Suspend
  else
    Result:= True;
end;

procedure TSyncIn.Stop;
begin
  if Suspended then
    Exit;
  FWaveIn.Stop;
  Suspend;
end;

procedure TSyncIn.SetAccTime(aTime: String);
var
  aCode: Integer;
  aVal: Double;
begin
  Val(aTime, aVal, aCode);
  FAccTime:= Round(aVal*1000);
end;

procedure TSyncIn.Validation(aVal: Boolean);
begin
  if not aVal then
    FAccInd:= 0;
  FWaveIn.Validation(aVal);
end;

procedure TSyncIn.OnlineProc(aData: TWaveData);
begin
end;

procedure TSyncIn.AsyncProc(aData: TWaveData);
begin
end;

procedure TSyncIn.AcquireData(aData: TWaveData);
var
  i, j, aTime: Integer;
begin
  OnlineProc(aData);
  if Length(aData[0]) <> Length(FAccData[0]) then begin
    SetWaveLength(FAccData, Length(aData[0]));
    FAccInd:= 0;
  end;
  if FAccInd = 0 then
    MoveWave(aData, FAccData)
  else
    for i:= 0 to High(aData[0]) do for j:= 0 to 1 do
      FAccData[j,i]:= FAccData[j,i] + aData[j,i];
  Inc(FAccInd);
  aTime:= DateTimeToTimeStamp(Now - FDateTime).Time;
  if aTime > 60000 then
    FAllowData:= True;
  if (aTime < FAccTime) or (not FAllowData) then begin
    if not FAccumulation then
      FAccInd:= 0;
    Exit;
  end;  
  FDateTime:= Now;
  if FAccInd > 1 then begin
    for i:= 0 to High(aData[0]) do for j:= 0 to 1 do
      FAccData[j,i]:= FAccData[j,i] / FAccInd;
    FAccInd:= 0;
  end;
  SendData(FAccData);
end;

procedure TSyncIn.SendData(aData: TWaveData);
begin
  SetWaveLength(FData, Length(aData[0]));
  MoveWave(aData, FData);
  FAllowData:= False;
  PulseEvent(FEvent);
end;

procedure TSyncIn.ExecEvent;
begin
  AsyncProc(FData);
  if Assigned(FWaveProc) then
    FWaveProc(FData);
end;

procedure TSyncIn.SetErrorMes(aErrorMes: TGetStrProc);
begin
  FWaveIn.ErrorMes:= aErrorMes;
end;

procedure TSyncIn.AllowData(Sender: TObject);
begin
  FAllowData:= True;
end;

end.
