unit uWaveOut;

interface

uses
  MMSystem, uWaveControl, uFFT;

type
  TWaveOut = class(TWaveControl)
  private
    FData: array[0..1] of TBufferData;
    FCurID, FAuxID: Integer;
    FResetData: Boolean;
    FValidate: TBooleanProc;
    FWaveData: TWaveData;
    procedure FillBuffer(aWaveHdr: pWAVEHDR);
    procedure SwapID;
    procedure OnResetData;
  protected
    procedure OnStop; override;
    procedure ExecData(aWaveHdr: pWAVEHDR); override;
    procedure WaveReset; override;
    procedure CloseWave; override;
  public
    constructor Create;
    destructor Destroy; override;
    function DataCount: Integer;
    function Start: Boolean;
    procedure Stop;
    procedure SetData(aWaveData: TWaveData);
    procedure SetRange(aMin, aMax, aAmp: Double; aCount: Integer);
    property Validate: TBooleanProc write FValidate;
  end;

procedure GenSpectrum(aWave: pWaveMask; aCount: Integer;
                      aMin, aMax: Integer; aAmp: Double);

implementation

uses
  Sysutils;

constructor TWaveOut.Create;
begin
  inherited;
  FCurID:= 0;
  FAuxID:= 1;
  FValidate:= nil;
  FWaveData:= cnNil2;
end;

destructor TWaveOut.Destroy;
begin
  inherited;
end;

function TWaveOut.Start: Boolean;
var
  i: Integer;
begin
  Result:= True;
  if not DoStart then
    Exit;
  FPosition:= 0;
  FResetData:= False;
  if waveOutOpen(@FDeviceHD, FDeviceID, @FFormatEx, FEvent, 0,
    CALLBACK_EVENT or WAVE_FORMAT_DIRECT or WAVE_ALLOWSYNC) <> MMSYSERR_NOERROR then begin
      Suspend;
      inherited OnStop;
      Result:= False;
      if Assigned(FErrorMes) then
        FErrorMes('Can not open output device ' + IntToStr(FDeviceID));
      Exit;  
    end;
  for i:= 0 to High(FWHdr) do begin
    waveOutPrepareHeader(FDeviceHD, @FWHdr[i], SizeOf(FWHdr[i]));
    FillBuffer(@FWHdr[i]);
    waveOutWrite(FDeviceHD, @FWHdr[i], SizeOf(FWHdr[i]));
  end;
end;

procedure TWaveOut.FillBuffer(aWaveHdr: pWAVEHDR);
var
  aMask: pDataMask;
  aMaskPosition, aMaskCount, aCount: Integer;
begin
  aMask:= pDataMask(aWaveHdr.lpData);
  aMaskCount:= aWaveHdr.dwBufferLength div 4;
  aMaskPosition:= 0;
  repeat
    aCount:= aMaskCount - aMaskPosition;
    if FPosition + aCount > Length(FData[FCurID]) then
      aCount:= Length(FData[FCurID]) - FPosition;
    Move(FData[FCurID,FPosition], aMask[aMaskPosition], aCount*4);
    Inc(aMaskPosition, aCount);
    Inc(FPosition, aCount);
    if FPosition = Length(FData[FCurID]) then
      FPosition:= 0;
  until aMaskPosition = aMaskCount;
end;

procedure TWaveOut.OnResetData;
begin
  SwapID;
  waveOutReset(FDeviceHD);
  FResetData:= False;
end;

procedure TWaveOut.ExecData(aWaveHdr: pWAVEHDR);
var
  aReset: Boolean;
begin
  if FResetData then begin
    aReset:= True;
    Synchronize(OnResetData);
  end else
    aReset:= False;
  FillBuffer(aWaveHdr);
  waveOutWrite(FDeviceHD, aWaveHdr, SizeOf(aWaveHdr^));
  if aReset and Assigned(FValidate) then
    FValidate(True);
end;

procedure TWaveOut.CloseWave;
begin
  waveOutClose(FDeviceHD);
  inherited;
end;

procedure TWaveOut.WaveReset;
begin
  waveOutReset(FDeviceHD);
end;

procedure TWaveOut.Stop;
begin
  if not DoStop then
    Exit;
  waveOutReset(FDeviceHD);
end;

procedure TWaveOut.OnStop;
var
  i: Integer;
begin
  for i:= 0 to High(FWHdr) do
    waveOutUnPrepareHeader(FDeviceHD, @FWHdr[i], SizeOf(FWHdr[i]));
  inherited;
end;

function TWaveOut.DataCount: Integer;
begin
  Result:= Length(FData[FCurID]);
end;

procedure TWaveOut.SetData(aWaveData: TWaveData);
const
  cRange = (High(SmallInt) - Low(SmallInt)) / 2;
var
  i, j: Integer;
begin
  if Length(FData[FAuxID]) <> Length(aWaveData[0]) then
    SetLength(FData[FAuxID], Length(aWaveData[0]));
  for i:= 0 to High(FData[FAuxID]) do
    for j:= 0 to 1 do
      FData[FAuxID,i,j]:= Trunc(cRange*aWaveData[j,i]);
  if Suspended then
    SwapID
  else begin
    if Assigned(FValidate) then
      FValidate(False);
    FResetData:= True;
  end;  
end;

procedure TWaveOut.SwapID;
var
  aID: Integer;
begin
  aID:= FAuxID;
  FAuxID:= FCurID;
  FCurID:= aID;
  FPosition:= 0;
end;

procedure GenSpectrum(aWave: pWaveMask; aCount: Integer;
                      aMin, aMax: Integer; aAmp: Double);
const
  cnHarmData: array[1..16,1..16] of LongWord = (
    (0,0,3,2,5,13,13,27,88,215,493,218,1781,3083,5086,4958),
    (0,0,0,6,4,4,22,26,52,88,88,474,332,3562,8042,16084),
    (0,0,0,0,12,8,16,32,52,88,284,860,948,2684,1484,16084),
    (0,0,0,0,0,24,32,16,88,104,208,832,944,1896,3792,4456),
    (0,0,0,0,0,0,48,64,32,176,208,208,704,3792,3792,7600),
    (0,0,0,0,0,0,0,96,64,64,352,416,704,3328,1696,7584),
    (0,0,0,0,0,0,0,0,192,256,128,704,832,1408,2624,3392),
    (0,0,0,0,0,0,0,0,0,384,512,256,1408,1664,7808,3328),
    (0,0,0,0,0,0,0,0,0,0,768,512,512,2816,3328,5632),
    (0,0,0,0,0,0,0,0,0,0,0,1536,1024,1024,5632,6656),
    (0,0,0,0,0,0,0,0,0,0,0,0,3072,4096,2048,11264),
    (0,0,0,0,0,0,0,0,0,0,0,0,0,6144,4096,4096),
    (0,0,0,0,0,0,0,0,0,0,0,0,0,0,12288,8192),
    (0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,24576),
    (0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0),
    (0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0));
var
  i: Integer;
  aA, aB, aMod, aMinA, aMaxA: Double;
begin
  if aMax <= 16 then begin
    FillChar(aWave[0], aCount*SizeOf(aWave[0]), 0);
    for i:= aMin to aMax do
      if (cnHarmData[aMin, aMax] and (1 shl (i-1)) = 0) then
        aWave[2*i]:=  1
      else
        aWave[2*i]:= -1;
    WaveFFT(aWave, aCount, False);
  end else begin
    aA:= Pi/(aCount);
    aB:= aA*(aMax-aMin)/aCount;
    aA:= 2*aA*aMin;
    for i:= 0 to aCount - 1 do begin
      aWave[i]:= Sin(i*(aA + aB*i));
    end;
    WaveFFT(aWave, aCount, True);
    for i:= 0 to (aCount div 2)-1 do begin
      if (i < aMin) or (i > aMax) then begin
        aWave[2*i]:= 0;
        aWave[2*i+1]:= 0;
      end else begin
        aMod:= Sqrt(Sqr(aWave[2*i])+Sqr(aWave[2*i+1]));
        aWave[2*i]:= aWave[2*i]/aMod;
        aWave[2*i+1]:= aWave[2*i+1]/aMod;
      end;
    end;
    WaveFFT(aWave, aCount, False);
  end;
  // Normalization
  aMinA:= aWave[0];
  aMaxA:= aMinA;
  for i:= 1 to aCount-1 do begin
    if aWave[i]> aMaxA then
      aMaxA:= aWave[i];
    if aWave[i]< aMinA then
      aMinA:= aWave[i];
  end;
  aA:= (aMaxA + aMinA)/2;
  aB:= 1.999*aAmp/(aMaxA - aMinA);
  for i:= 0 to aCount-1 do
    aWave[i]:= (aWave[i] - aA)*aB;
end;

procedure TWaveOut.SetRange(aMin, aMax, aAmp: Double; aCount: Integer);
var
  aMinHarm, aMaxHarm: Integer;
begin
  if aCount <> Length(FWaveData[0]) then
    SetLength(FWaveData[0], aCount);
  aMinHarm:= GetHarmonic(aMin, Length(FWaveData[0]));
  aMaxHarm:= GetHarmonic(aMax, Length(FWaveData[0]));
  GenSpectrum(@FWaveData[0,0], Length(FWaveData[0]), aMinHarm, aMaxHarm, aAmp);
  FWaveData[1]:= FWaveData[0];
  SetData(FWaveData);
end;

end.
